import asyncio
from asyncio.transports import DatagramTransport
from contextlib import contextmanager
from random import random
from typing import Any, Awaitable, Iterator, Optional, TypeVar

from aiodogstatsd import protocol, typedefs
from aiodogstatsd.compat import get_event_loop

__all__ = ("Client",)

_T = TypeVar("_T")


class Client:
    __slots__ = (
        "_host",
        "_port",
        "_namespace",
        "_constant_tags",
        "_state",
        "_protocol",
        "_pending_queue",
        "_pending_queue_size",
        "_listen_future",
        "_listen_future_join",
        "_read_timeout",
        "_close_timeout",
        "_sample_rate",
    )

    @property
    def connected(self) -> bool:
        return self._state == typedefs.CState.CONNECTED

    @property
    def closing(self) -> bool:
        return self._state == typedefs.CState.CLOSING

    @property
    def disconnected(self) -> bool:
        return self._state == typedefs.CState.DISCONNECTED

    def __init__(
        self,
        *,
        host: str = "localhost",
        port: int = 9125,
        namespace: Optional[typedefs.MNamespace] = None,
        constant_tags: Optional[typedefs.MTags] = None,
        read_timeout: float = 0.5,
        close_timeout: Optional[float] = None,
        sample_rate: typedefs.MSampleRate = 1,
        pending_queue_size: int = 2 ** 16,
    ) -> None:
        """
        Initialize a client object.

        You can pass `host` and `port` of the DogStatsD server, `namespace` to prefix
        all metric names, `constant_tags` to attach to all metrics.

        Also, you can specify: `read_timeout` which will be used to read messages from
        an AsyncIO queue; `close_timeout` which will be used as wait time for client
        closing; `sample_rate` can be used for adjusting the frequency of stats sending.
        """
        self._host = host
        self._port = port
        self._namespace = namespace
        self._constant_tags = constant_tags or {}

        self._state = typedefs.CState.DISCONNECTED

        self._protocol = DatagramProtocol()

        self._pending_queue: asyncio.Queue
        self._pending_queue_size = pending_queue_size

        self._listen_future: asyncio.Future
        self._listen_future_join: asyncio.Future

        self._read_timeout = read_timeout
        self._close_timeout = close_timeout
        self._sample_rate = sample_rate

    async def __aenter__(self) -> "Client":
        await self.connect()
        return self

    async def __aexit__(self, *args) -> None:
        await self.close()

    async def connect(self) -> None:
        loop = get_event_loop()
        await loop.create_datagram_endpoint(
            lambda: self._protocol, remote_addr=(self._host, self._port)
        )

        self._pending_queue = asyncio.Queue(maxsize=self._pending_queue_size)
        self._listen_future = asyncio.ensure_future(self._listen())
        self._listen_future_join = asyncio.Future()

        self._state = typedefs.CState.CONNECTED

    async def close(self) -> None:
        self._state = typedefs.CState.CLOSING

        try:
            await asyncio.wait_for(self._close(), timeout=self._close_timeout)
        except asyncio.TimeoutError:
            pass

        self._state = typedefs.CState.DISCONNECTED

    async def _close(self) -> None:
        await self._listen_future_join
        self._listen_future.cancel()

        await self._protocol.close()

    def gauge(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: Optional[typedefs.MSampleRate] = None,
    ) -> None:
        """
        Record the value of a gauge, optionally setting tags and a sample rate.
        """
        self._report(name, typedefs.MType.GAUGE, value, tags, sample_rate)

    def increment(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue = 1,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: Optional[typedefs.MSampleRate] = None,
    ) -> None:
        """
        Increment a counter, optionally setting a value, tags and a sample rate.
        """
        self._report(name, typedefs.MType.COUNTER, value, tags, sample_rate)

    def decrement(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue = 1,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: Optional[typedefs.MSampleRate] = None,
    ) -> None:
        """
        Decrement a counter, optionally setting a value, tags and a sample rate.
        """
        value = -value if value else value
        self._report(name, typedefs.MType.COUNTER, value, tags, sample_rate)

    def histogram(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: Optional[typedefs.MSampleRate] = None,
    ) -> None:
        """
        Sample a histogram value, optionally setting tags and a sample rate.
        """
        self._report(name, typedefs.MType.HISTOGRAM, value, tags, sample_rate)

    def distribution(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: Optional[typedefs.MSampleRate] = None,
    ) -> None:
        """
        Send a global distribution value, optionally setting tags and a sample rate.
        """
        self._report(name, typedefs.MType.DISTRIBUTION, value, tags, sample_rate)

    def timing(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: Optional[typedefs.MSampleRate] = None,
    ) -> None:
        """
        Record a timing, optionally setting tags and a sample rate.
        """
        self._report(name, typedefs.MType.TIMING, value, tags, sample_rate)

    async def _listen(self) -> None:
        try:
            while self.connected:
                await self._listen_and_send()
        finally:
            # Note that `asyncio.CancelledError` raised on app clean up
            # Try to send remaining enqueued metrics if any
            while not self._pending_queue.empty():
                await self._listen_and_send()
            self._listen_future_join.set_result(True)

    async def _listen_and_send(self) -> None:
        coro = self._pending_queue.get()

        try:
            buf = await asyncio.wait_for(coro, timeout=self._read_timeout)
        except asyncio.TimeoutError:
            pass
        else:
            self._protocol.send(buf)

    def _report(
        self,
        name: typedefs.MName,
        type_: typedefs.MType,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: Optional[typedefs.MSampleRate] = None,
    ) -> None:
        # Ignore any new incoming metric if client in closing or disconnected state
        if self.closing or self.disconnected:
            return

        sample_rate = sample_rate or self._sample_rate
        if sample_rate != 1 and random() > sample_rate:
            return

        # Resolve full tags list
        all_tags = dict(self._constant_tags, **tags or {})

        # Build metric
        metric = protocol.build(
            name=name,
            namespace=self._namespace,
            value=value,
            type_=type_,
            tags=all_tags,
            sample_rate=sample_rate,
        )

        # Enqueue metric
        try:
            self._pending_queue.put_nowait(metric)
        except asyncio.QueueFull:
            pass

    @contextmanager
    def timeit(
        self,
        name: typedefs.MName,
        *,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: Optional[typedefs.MSampleRate] = None,
        threshold_ms: Optional[typedefs.MValue] = None,
    ) -> Iterator[None]:
        """
        Context manager for easily timing methods.
        """
        loop = get_event_loop()
        started_at = loop.time()

        try:
            yield
        finally:
            value = (loop.time() - started_at) * 1000
            if not threshold_ms or value > threshold_ms:
                self.timing(name, value=int(value), tags=tags, sample_rate=sample_rate)

    def timeit_task(
        self,
        coro: Awaitable[_T],
        name: typedefs.MName,
        *,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: Optional[typedefs.MSampleRate] = None,
        threshold_ms: Optional[typedefs.MValue] = None,
    ) -> "asyncio.Task[_T]":
        """
        Creates a task and returns it, adds a done callback for sending time metric when
        done and if exceeds threshold.
        """
        loop = get_event_loop()
        started_at = loop.time()

        def _callback(_: Any) -> None:
            duration = (loop.time() - started_at) * 1000
            if threshold_ms and duration < threshold_ms:
                return
            self.timing(name, value=int(duration), tags=tags, sample_rate=sample_rate)

        task = loop.create_task(coro)
        task.add_done_callback(_callback)
        return task


class DatagramProtocol(asyncio.DatagramProtocol):
    __slots__ = ("_transport", "_closed")

    def __init__(self) -> None:
        self._transport: Optional[DatagramTransport] = None
        self._closed: asyncio.Future

    async def close(self) -> None:
        if self._transport is None:
            return

        self._transport.close()
        await self._closed

    def connection_made(self, transport):
        self._transport = transport
        self._closed = asyncio.Future()

    def connection_lost(self, _exc):
        self._transport = None
        self._closed.set_result(True)

    def send(self, data: bytes) -> None:
        if self._transport is None:
            return

        try:
            self._transport.sendto(data)
        except Exception:
            # Errors should fail silently so they don't affect anything else
            pass



import enum
from typing import Mapping, Union

__all__ = (
    "CState",
    "MName",
    "MNamespace",
    "MType",
    "MValue",
    "MSampleRate",
    "MTagKey",
    "MTagValue",
    "MTags",
)


MName = str
MNamespace = str
MValue = Union[float, int]
MSampleRate = Union[float, int]

MTagKey = str
MTagValue = Union[float, int, str]
MTags = Mapping[MTagKey, MTagValue]


@enum.unique
class MType(enum.Enum):
    COUNTER = "c"
    DISTRIBUTION = "d"
    GAUGE = "g"
    HISTOGRAM = "h"
    TIMING = "ms"


@enum.unique
class CState(enum.IntEnum):
    CONNECTED = enum.auto()
    CLOSING = enum.auto()
    DISCONNECTED = enum.auto()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiodogstatsd', 'aiodogstatsd.contrib']

package_data = \
{'': ['*']}

extras_require = \
{'aiohttp': ['aiohttp>=3.0.0'],
 'sanic': ['sanic>=20.3.0'],
 'starlette': ['starlette>=0.13.0']}

setup_kwargs = {
    'name': 'aiodogstatsd',
    'version': '0.15.0',
    'description': 'An asyncio-based client for sending metrics to StatsD with support of DogStatsD extension',
    'long_description': '# aiodogstatsd\n\n[![Build Status](https://github.com/Gr1N/aiodogstatsd/workflows/default/badge.svg)](https://github.com/Gr1N/aiodogstatsd/actions?query=workflow%3Adefault) [![codecov](https://codecov.io/gh/Gr1N/aiodogstatsd/branch/master/graph/badge.svg)](https://codecov.io/gh/Gr1N/aiodogstatsd) ![PyPI](https://img.shields.io/pypi/v/aiodogstatsd.svg?label=pypi%20version) ![PyPI - Downloads](https://img.shields.io/pypi/dm/aiodogstatsd.svg?label=pypi%20downloads) ![GitHub](https://img.shields.io/github/license/Gr1N/aiodogstatsd.svg)\n\nAn asyncio-based client for sending metrics to StatsD with support of [DogStatsD](https://docs.datadoghq.com/developers/dogstatsd/) extension.\n\nLibrary fully tested with [statsd_exporter](https://github.com/prometheus/statsd_exporter) and supports `gauge`, `counter`, `histogram`, `distribution` and `timing` types.\n\n`aiodogstatsd` client by default uses _9125_ port. It\'s a default port for [statsd_exporter](https://github.com/prometheus/statsd_exporter) and it\'s different from _8125_ which is used by default in StatsD and [DataDog](https://www.datadoghq.com/). Initialize the client with the proper port you need if it\'s different from _9125_.\n\n## Installation\n\nJust type:\n\n```sh\n$ pip install aiodogstatsd\n```\n\n## At a glance\n\nJust simply use client as a context manager and send any metric you want:\n\n```python\nimport asyncio\n\nimport aiodogstatsd\n\n\nasync def main():\n    async with aiodogstatsd.Client() as client:\n        client.increment("users.online")\n\n\nasyncio.run(main())\n```\n\nPlease follow [documentation](https://gr1n.github.io/aiodogstatsd) or look at [`examples/`](https://github.com/Gr1N/aiodogstatsd/tree/master/examples) directory to find more examples of library usage, e.g. integration with [`AIOHTTP`](https://aiohttp.readthedocs.io/), [`Sanic`](https://sanicframework.org/) or [`Starlette`](https://www.starlette.io) frameworks.\n\n## Contributing\n\nTo work on the `aiodogstatsd` codebase, you\'ll want to clone the project locally and install the required dependencies via [poetry](https://poetry.eustace.io):\n\n```sh\n$ git clone git@github.com:Gr1N/aiodogstatsd.git\n$ make install\n```\n\nTo run tests and linters use command below:\n\n```sh\n$ make lint && make test\n```\n\nIf you want to run only tests or linters you can explicitly specify which test environment you want to run, e.g.:\n\n```sh\n$ make lint-black\n```\n\n## License\n\n`aiodogstatsd` is licensed under the MIT license. See the license file for details.\n',
    'author': 'Nikita Grishko',
    'author_email': 'gr1n@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Gr1N/aiodogstatsd',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import asyncio
import sys

__all__ = ("get_event_loop",)


def _get_event_loop_factory():  # pragma: no cover
    if sys.version_info >= (3, 7):
        return asyncio.get_running_loop  # type: ignore

    return asyncio.get_event_loop


get_event_loop = _get_event_loop_factory()
